/*  ___   ___  ___  _   _  ___   ___   ____ ___  ____  
 * / _ \ /___)/ _ \| | | |/ _ \ / _ \ / ___) _ \|    \ 
 *| |_| |___ | |_| | |_| | |_| | |_| ( (__| |_| | | | |
 * \___/(___/ \___/ \__  |\___/ \___(_)____)___/|_|_|_|
 *                  (____/ 
 * Arduino Mecanum Omni Direction Wheel Robot Car
 * Tutorial URL https://osoyoo.com/?p=58245
 * CopyRight www.osoyoo.com
 */
#define EC1_A 50
#define EC1_B 51
#define EC2_A 52
#define EC2_B 53
#define speedPinR 9   //  Front Wheel PWM pin connect Right MODEL-X ENA 
#define RightMotorDirPin1  22    //Front Right Motor direction pin 1 to Right MODEL-X IN1  (K1)
#define RightMotorDirPin2  24   //Front Right Motor direction pin 2 to Right MODEL-X IN2   (K1)                                 
#define LeftMotorDirPin1  26    //Front Left Motor direction pin 1 to Right MODEL-X IN3 (K3)
#define LeftMotorDirPin2  28   //Front Left Motor direction pin 2 to Right MODEL-X IN4 (K3)
#define speedPinL 10   //  Front Wheel PWM pin connect Right MODEL-X ENB

int counter1 = 0;
int counter2 = 0;
int a1State;
int a1LastState;
int a2State;
int a2LastState;

// PID variables
double kp = 0.5, ki = 0.05, kd = 0.01;  // Adjusted PID constants
double targetSpeed = 120;  // Default left motor speed
double rightMotorSpeed = 120;  // Start with the same speed
double error = 0, previousError = 0, integral = 0, derivative = 0;
unsigned long previousTime = 0;

void setup() {
  pinMode(EC1_A, INPUT);
  pinMode(EC1_B, INPUT);
  pinMode(EC2_A, INPUT);
  pinMode(EC2_B, INPUT);
  pinMode(RightMotorDirPin1,OUTPUT);
  pinMode(RightMotorDirPin2,OUTPUT);
  pinMode(LeftMotorDirPin1,OUTPUT);
  pinMode(LeftMotorDirPin2,OUTPUT);
  pinMode(speedPinL, OUTPUT);  
  pinMode(speedPinR, OUTPUT);  

  Serial.begin(9600);

  // Reads the initial state of the encoders
  a1LastState = digitalRead(EC1_A);
  a2LastState = digitalRead(EC2_A);

  // Initialize motors
  leftmotor(120);  // Set left motor to default speed
  rightmotor(120); // Initially set right motor to the same speed
}

void loop() {
  a1State = digitalRead(EC1_A);  // Read the current state of EC1_A
  a2State = digitalRead(EC2_A);  // Read the current state of EC2_A

  // Update encoder counts
  if (a1State != a1LastState) {
    if (digitalRead(EC1_B) != a1State) {
      counter1++;
    } else {
      counter1--;
    }
 //   Serial.print("Position1: ");
  //  Serial.println(counter1);
  }
  if (a2State != a2LastState) {
    if (digitalRead(EC2_B) != a2State) {
      counter2++;
    } else {
      counter2--;
    }
  //  Serial.print("Position2: ");
  //  Serial.println(counter2);
  }

  // Calculate motor speeds based on encoder counts
  double leftSpeed = counter1;  // Speed of left motor
  double rightSpeed = counter2; // Speed of right motor

  // PID control to keep right motor in sync with left motor
  unsigned long currentTime = millis();
  double elapsedTime = (currentTime - previousTime) / 1000.0;
  previousTime = currentTime;

  // PID calculations
  error = leftSpeed - rightSpeed;

  // Add a small dead zone to avoid overcorrection
  if (abs(error) < 5) {
    error = 0;
  }

  integral += error * elapsedTime;
  derivative = (error - previousError) / elapsedTime;

  // Calculate the PID output (adjust right motor speed)
  double output = kp * error + ki * integral + kd * derivative;

  // Adjust right motor speed based on PID output
  rightMotorSpeed = constrain(120 + output, 30, 255);  // Constrain the speed between 30 and 255 to avoid stopping
  rightmotor(rightMotorSpeed);

  // Debugging output
  Serial.print("Left Speed: ");
  Serial.print(leftSpeed);
  Serial.print("Right Speed: ");
  Serial.print(rightSpeed);
  Serial.print("Right Motor PWM: ");
  Serial.println(rightMotorSpeed);

  previousError = error;  // Update error for next iteration

  // Update previous states for encoders
  a1LastState = a1State;
  a2LastState = a2State;

  delay(100);  // Small delay to smooth out loop
}

void leftmotor(int speed)
{
  digitalWrite(LeftMotorDirPin1,HIGH);
  digitalWrite(LeftMotorDirPin2,LOW); 
  analogWrite(speedPinL,speed);
}
void rightmotor(int speed)
{
  digitalWrite(RightMotorDirPin1,HIGH);
  digitalWrite(RightMotorDirPin2,LOW); 
  analogWrite(speedPinR,speed);
}
